( 3iPort.fs - words to read/write to ports in binary )
\ Assad Ebrahim
\ 2020-09-09---Wed---01:31 - started testing the ICs in Forth
\ 2020-09-12---Sat---23:34 - finished 

\ Arduino Nano with 3iForth.avr
\ Hardware 1: Built 4-port INPUT to 74154 IC (4-bits to 16-line decoder)
\    G1 control signal on chip pulled low permanently
\    2345 on Nano connecting to ABCD on chip (A=lsb, D = msb), 
\    6 is control pin connected to G2 on chip, pull low to start the decoding
\   LEDs connected via 860R resistor from VCC (5V) to address pin of chip (0-12) so they light when the chip lines are pulled low
\ Hardware 2: Built 4-port OUTPUT from 2x cascaded 74148 ICs and a 7408 (AND) IC into a 16-line to 4-bits priority encoder

\ Usage: help-port
\ The below code writes an address/channel (0-15) coded as 4-bits to a 4-bit port and uses 74154 IC to decode and turn on LED with associated address.  The use of the chip saves consumption of 16 I/O pins instead requiring 5, 4 for 4-bit port, 1 for control line


\ PIN LEVEL INSTRUCTIONS FROM AVR and VIRTUAL PORT EMULATION FROM FORTH.

\ WRITING / OUTPUT TO PORT

: out4  5 4 3 2 ;    \ pins for output port, lsb top of stack
: in4   11 10 9 8 ;  \ pins for input port, lsb on top
6  constant ctrl-out \ control signal, active low
12 constant ctrl-in
: out4-set  out4 out out out out ctrl-out out ;
: in4-set    in4  in in in in ctrl-in in ;

: p> 20 ms  ctrl-out hi ;  \ decode-stop: all commands are now buffered, i.e. the uC reflects the pins but the IC does not ref
: >p 20 ms  ctrl-out lo ;  \ decode-start: all commands are now real-time reflected by the IC

: [bit>p] ( 1/0 pin -- ) swap if hi else lo then ;  \ active high signal (inputs to 74154)
: bit>p   ( n pinmask -- ) and swap [bit>p] ;
: >>p   ( n -- )  
    p> \ hold port
    >r out4 
    r@ $1 bit>p
    r@ $2 bit>p
    r@ $4 bit>p
    r> $8 bit>p
    >p \ show
;

: port-test 13 0 do i >>p loop p> ;  \ auto validation

: keys-test  \ piano application validation
    \ white keys
    0  >>p 
    2  >>p
    4  >>p
    5  >>p
    7  >>p
    9  >>p
   11  >>p
   12  >>p
    \ black keys
    1  >>p
    3  >>p
    6  >>p
    8  >>p
   10  >>p
   p>
;


\ READING INPUT FROM PORT

\ Hardware: 
\ Arduino Nano with 3iForth.avr
\ connected to 2x 74148 ICs cascaded with 7408 (AND), with pins 9,10,11,12 as data pins (lsb-msb) and pin 8 as enable (pull low)

\ Usage: 
\ port4 port4-in   setup ports on Arduino
\ port4read-test   continuously reads the 4-bit input port.  Test by setting 0-7 active low on 74148#1 then 8-15 low on 74148#2
\ Ctrl-C to exit the infinite loop

: alow ( 0/1 -- 1/0 ) 1 + 2 mod ;  \ active low
\ for each individual 74148 IC (8 line to 3 bit priority encoder)
: port3A 10 11 12 9 ;
: port3B 15 16 17 14 ;
: port3-in ( port -- ) in in in in ;
: 3drop 2drop drop ;
: read3 ( port -- lsb0 1 msb2 ) readpin alow if rot readpin alow -rot swap readpin alow swap readpin alow else 3drop then ;
: read3A port3A read3 ;
: read3B port3B read3 ;
: 3.  . . . ; 

\ for cascaded 2x 74148 ICs into 7408 AND chip (16 line to 4 bit priority encoder)
: port4 ( -- lsb ... msb valid )8 9 10 11  12 ;  \ pin12 = data valid pin (valid if active low), pins 8-11 are 4-bit port read
: port4-in ( port -- ) port4 in in in in in ;
: [read-port4] ( msb ... lsb -- )  4 0 do readpin alow . loop ;
: read4 ( port -- ) readpin alow if  [read-port4] else 4drop then ;

: port4-read  cr begin port4 readpin alow if [read-port4] cr else 4 drop then 0 until ;

: read4 
cr
decimal
11 pinread .
10 pinread .
9  pinread .
8  pinread .
;

: read4low
cr
decimal
11 pinread not .
10 pinread not .
9  pinread not .
8  pinread not .
;




: help-port
   cr cr
   cr ." PORT HELP"
   cr ." Dependencies: 3iForth.avr, 3iForth.fs, 3iLight.fs, 3iPort.fs"
   cr 
   cr ." 4-bit port OUT using 74154 IC (4-bit to 16 line decoder)"
   cr ." Active high inputs, active low outputs."
   cr 
   cr ." out4-set - set pins 2,3,4,5,6 as 4-port OUT and control pin"
   cr ." n >>p  - send number to port"
   cr ." p>   - port off"
   cr ." >p   - port on"
   cr ." port-test    keys-test    - demonstrate functionality"
   cr
   cr ." 4-bit port IN using 2x 74148 IC cascaded into 7408 AND chip (16-line to 4-bit priority encoder)"
   cr ." Active low inputs and active low outputs."
   cr ." port4-in - set pins 8,9,10,11,12 as 4-port IN and data valid pin"
   cr ." port4read-test - reads port printing encoded inputs until Ctrl-C pressed"
   cr
;

help-port
