\ 3iForth.fs -- 3-Instruction Forth -- HOST CODE (GForth on PC platform) targeting ATMEL 328 on Arduino Nano
\ Assad Ebrahim, http://www.mathscitech.org/articles, assadebrahim2000@gmail.com
\ 2020-06-22

\ This program runs in GForth on Windows and communicates with Arduino Nano (Atmel 328 chip) 
\ Its communication partner is 3iForth.c compiled and uploaded to the Nano with the free Arduino IDE and default bootloader.
\ Upon startup or reset, the Nano sets up serial port (9600 baud) and waits for instructions over the serial line...
\ USAGE:  
\   1. Connect Arduino Nano to your laptop using USB mini cable
\   2. include serial.fs
\ Demonstration words are intended for the demo board built with 11 LEDs (6 red and 5 yellow) and a tiny speaker.
\ You can see this demonstration on the website @ http://www.mathscitech.org/articles/3iForth

\ c. 85 lines of Forth code

( * SERIAL PORT MANAGEMENT * )
variable handle   \ COM port file handle
: save-handle ( handle -- ) handle ! ;
: fh ( -- handle ) handle @ ;
: com-file ( n -- c-adr cnt )      \ builds COM string.  Test: 6 com-file type
    >r s" comX" 2dup + 1 - '0' r> + swap c! ; 
: com-open? ( handle errcode -- )
    ." ..."
    dup 0= if drop 1500 ms save-handle cr ." COM port opened.  OK." cr \ Note: 1000ms doesn't work, 1500ms works settling time
    else cr ." **error: check device is plugged into correct COM port" cr throw
    then ;
:  [OPEN-COM] ( n --  )         \ open COM port in binary mode.   USAGE:  6 open-com constant 
    com-file r/w bin open-file com-open? ;
:  CLOSE-COM ( -- )
    fh close-file throw cr ." COM port closed." cr ;


( * MEMORY MANAGEMENT * )
: bytes 1 * ;
create rcv 255 bytes allot    \ read buffer
:  c@+ ( adr -- adr' val ) 
    dup 1 + swap c@ ;
:  c!+ ( val adr -- adr' )
    dup 1 + -rot c! ;

( * SERIAL READ * )
: show-rcv ( cnt -- ) 
    rcv swap dup 0> if 0 cr ." <rcv: " do c@+ . ." " loop else drop then drop ;    \ display change to hex. to show hex values
: read-com ( -- cnt ) 
    rcv 50 fh read-file throw ( -- cnt )
    dup 0= if ." No data on port: " then dup .
    ;
: open-com ( n -- )
    [open-com] read-com show-rcv ;


( * HOST TO TARGET MESSAGES * )
create msg 20 bytes allot     \  write buffer
: msglen ( adr' -- n )  
    msg - ;
: show-msg ( msg cnt -- )                   \ USAGE: msg 10 show-msg
    0 cr ." msg>: "do c@+ . ." " loop drop ;          \ change . to hex. to show hex values
: tx  ( adr'  -- )                          \ transmit message and listen for response
    msg swap msglen   \   uncomment to show tx messages  2dup show-msg  
    fh write-file throw fh flush-file throw  \ transmit
    ;
: tx-rx ( -- ) 80 ms read-com show-rcv ;     \ 80 ms works, 60 intermittent, 25 doesn't work

: test1-tx ( -- )  0 6 9 1 1 1 msg c!+ c!+ c!+ c!+ c!+ c!+   msg 6 show-msg tx ;
: test2-tx ( -- )  1 6 9 1 1 1 msg c!+ c!+ c!+ c!+ c!+ c!+   msg 6 show-msg tx ;
: test3-tx ( -- )  1 6 9 1 0 1 msg c!+ c!+ c!+ c!+ c!+ c!+   msg 6 show-msg tx ;
: test-tx ( -- ) test1-tx test2-tx ;

( * REMOTE INSTRUCTIONS TO TARGET - there are only FOUR that work with parameters on the target data stack * )
\ Notes: 
\ 1.   tx ;      only xcall eventually transits.  the others prepare the msg buffer (fewer comms -> more efficient)
\ 2.   -- adr'   each function accepts an address and provides the next address for easy chaining into the msg buffer
: xstore  ( val adr -- adr' )  1 swap c!+ c!+ ;
: x_s     ( adr -- adr' )      2 swap c!+ ;
: xdrop   ( adr -- adr' )      3 swap c!+ ;
: xcall   ( adr tok -- adr' )  6 rot c!+ c!+ tx ;

( * HOST INTERFACE  CALL - on the host machine we don't need to know the mechanism, we think we live in the target and call the exported functions directly using Forth * )
: xstore2 ( a b -- adr' )
    swap msg xstore xstore ;
: pinMode ( mode pin -- )    \ ex: 1 9 pinmode  (output)
    xstore2 0 xcall 60 ms ; \ Note: need settle time on pin when changing mode.  60 ms is magic number
: pinset ( level pin -- )  \ ex: 1 9 pinset   0 9 pinset
    xstore2 1 xcall ;
: digitalWrite pinset ;
\ : delayms ( ms -- )  \ ex: 128 delayms   NOT NEEDED on TARGET as have it on hose
\    msg xstore 2 xcall ; 
\ : delay delayms ;


( * PROBLEM ORIENTED LANGUAGE --- setting up a simplified yet powerful custom dictionary * )

: high ( pin -- )  1 swap pinset ;
: low ( pin -- )   0 swap pinset ;
: lo low ;
: hi high ;
: out ( pin -- )   1 swap pinmode ;
: in ( pin -- )    0 swap pinmode ;
: testpin ( n -- )  dup dup out high 500 ms low ." done." ;

: reload ( -- )   \ ensure com port is closed before reloading which clobbers fh
  fh 0> if close-com then
  s" serial.fs" included ;

: help
    cr cr cr cr cr
    cr ."  ------------------------------------------------------------"
    cr ."  Welcome - your Atmel 328 chip is ready for you to instruct:"
    cr ."  1.  6 open-com    -  opens communication to port 6 "    
    cr ."  5.  all-out       - sets all pins to out"
    cr ."  3.  2 high        - sends voltage to pin 2  (choose pins 2 through 12)"
    cr ."  4.  2 low         - sends voltage to pin 2 (choose pins 2 through 12)"
    cr ."  5.  close-com"
    cr ."  3 rules: Try things out (you can't break anything)!  Have fun!  Make something cool!"
    cr ."  1 tip: Ask for help just before you get so frustrated you want to quit!"
    cr cr
;

\ ****************************
help

\ ================================

\ EXERCISES: 

: all-out ( -- ) cr ." ..." cr 13 2 do i out loop ;
: all-high 13 2 do i high loop ;
: all-low ( -- ) 13 2 do i low loop ;
\ Solution 1 - brute force
\ : yellows-high 3 high 5 high 7 high 9 high 11 high ;
\ : yellows-low 3 low 5 low 7 low 9 low 11 low ;
\ Solution 2 - loop
\ : yellows-high ( -- )   3 5 7 9 11  5 0 do high loop ; 
\ : yellows-low ( -- )   3 5 7 9 11  5 0 do low loop ; 
\ Solution 3 - factor
: yellows 3 5 7 9 11 ;
: yellows-high ( -- )   yellows 5 0 do high loop ; 
: yellows-low ( -- )   yellows  5 0 do low loop ; 

: reds 2 4 6 8 10 12 ;
: reds-high reds 6 0 do hi loop ;
: reds-low reds 6 0 do low loop ;

: flash all-high 200 ms all-low 200 ms ;
: n-flash ( n -- ) 0 do flash loop ;  \ ex: 10 n-flash
: flash-yellows   yellows-high 200 ms yellows-low 200 ms ;
: flash-reds      reds-high 200 ms reds-low 200 ms ;
: n-bounce ( n -- ) 0 do flash flash flash-yellows flash-reds flash-yellows flash-reds flash loop ;

: kit-delay 25 ms ;  \ delay for successive changes to look smooth
: l2r-high 13 2 do i hi kit-delay loop ;
: l2r-low  13 2 do i low kit-delay loop ;
: r2l-high 13 2 do 14 i - hi kit-delay loop ;  \ note: only gets to 12
: r2l-low 13 2 do 14 i - low kit-delay loop ;
: n-kit  ( n -- ) 0 do l2r-high l2r-low r2l-high r2l-low loop ;

\ knight-rider effect
: l2r-vivid 2 high 3 high 13 4 do kit-delay i 2 - low i high loop ; 
: r2l-vivid 12 high 11 high 13 4 do kit-delay 16 i - low 16 i - 2 - high loop ; 
: vivid l2r-vivid 60 ms r2l-vivid 60 ms ;  \ a little stickier feel at the ends
: n-vivid ( n -- ) 0 do vivid loop all-low ;
